"""
A variable is a representation of a quantity with a variable set that 
defines a domain over which a function is valid. It therefore needs a name, 
variable type, and variable set. 
"""

#-------------------------------------------------------------------------------
import numpy as np
import collections
from probayes.icon import Icon, isiconic
from probayes.vtypes import eval_vtype, isunitset, isscalar, \
                        revtype, uniform, VTYPES, OO, OO_TO_NP
from probayes.variable_utils import parse_as_str_dict
from probayes.pscales import log_prob
from probayes.expression import Expression
from probayes.distribution import Distribution
from probayes.ops import and_op, or_op

# Defaults
DEFAULT_VNAME = 'var'
DEFAULT_VTYPE = bool

# Defult vsets by vtype
DEFAULT_VSETS = {bool: [False, True],
                  int: [0, 1], 
                float: [(-OO,), (OO,)]}

#-------------------------------------------------------------------------------
class Variable (Icon):
  """ A variable is a representation of a quantity with a variable set that 
  defines a domain over which a function is valid. It therefore needs a name, 
  variable type, and variable set. 
  
  While this class does not support defining probability density functions 
  (use RV for that), it does include an option to specify a univariate
  function that can be used for example to specify a monotonic variable 
  transformation:

  :example:
  >>> import numpy as np
  >>> import probayes as pb
  >>> x = pb.Variable('x', vtype=float)
  >>> x.set_ufun((np.exp, np.log))
  >>> print(x.ufun[0](1))
  2.718281828459045
  >>> print(x.ufun[-1](1))
  0.0
  >>> print(x.vset)
  [(-oo,), (oo,)]
  >>> print(x.vlims)
  [-inf  inf]
  >>> print(x.ulims)
  [ 0. inf]
  """
                    
  # Protected       
  _Delta = None      # A namedtuple generator for delta operations
  _vtype = None      # Variable type (bool, int, or float)
  _vset = None       # Variable set (array or 2-length tuple range)
  _vlims = None      # Numpy array of bounds of vset
  _ufun = None       # Univariate function for variable transformation
  _ulims = None      # self._vlims if not no_ucov else transformed self._vlims
  _isfinite = None   # all(isfinite(ulims))
  _length = None     # Difference in self._ulims
  _lhv = None        # Log hypervolume
  _nlhv = None       # Negative log hypervolume - always of transformed space
  _inside = None     # Lambda function for defining inside vset
  _delta = None      # Default delta operation
  _delta_args = None # Optional delta arguments 
  _delta_kwds = None # Optional delta keywords 
  _is_stochastic = False # Flag of whether stochastic
  _dexpr = None      # Deterministic expression

  # Private       
  __no_ucov = None   # Boolean flag to denote no univariate change of variables

#-------------------------------------------------------------------------------
  @property
  def is_stochastic(self):
    return self._is_stochastic

  def __init__(self, name=None,
                     vset=None, 
                     vtype=None,
                     *args,
                     **kwds):
    """ Initialiser sets name, vset, and ufun:

    :param name: Name of the variable - string as valid identifier.
    :param vset: variable set over which variable domain defined.
    :param vtype: variable type (bool, int, or float).
    :param *args: optional arguments to pass onto symbol representation.
    :param **kwds: optional keywords to pass onto symbol representation.

    Every Variable instance offers a factory function for delta specifications:

    :example:
    >>> import numpy as np
    >>> import probayes as pb
    >>> x = pb.Variable('x', vtype=float)
    >>> dx = x.Delta(0.5)
    >>> print(x.apply_delta(1.5, dx))
    2.0
    """
    self.name = name

    # Allow positions of vtype and vset to be swapped
    if isinstance(vtype, (list, tuple, set, range, np.ndarray)):
      vtype, vset = vset, vtype

    # If vtype and/or vset is/are specified, set it/them
    if vtype or vset:
      if vtype:
        self.vtype = vtype
      if vset:
        self.vset = vset
    else:
      self.vtype = vtype

    # Setting icon comes afterwards to allow passing vtype/vset assumptions
    self.set_icon(self.name, *args, **kwds)

    # Default delta to nothing
    self.set_delta()

#-------------------------------------------------------------------------------
  @property
  def name(self):
    return self._name

  @property
  def Delta(self):
    return self._Delta

  @name.setter
  def name(self, name=DEFAULT_VNAME):
    self._name = name
    self._Delta = collections.namedtuple('ð', [self._name])


#-------------------------------------------------------------------------------
  @property
  def vset(self):
    """ Property vset is the set over which the valid values are defined.
    :param vset: variable set over which domain defined (defaulted by vtype:
                 if bool:  vset = {False, True})
                 if int:   vset = {0, 1})
                 if float: vset = (-OO, OO)

    For non-float vtypes, vset may be a list, set, range, or NumPy array.

    For float vtypes, vset represents limits in the form:

    [lower, upper] - inclusive of both lower of upper values
    [(lower), upper] - exclusive of lower and inclusive of upper.
    [lower, (upper)] - inclusive of lower and exclusive of upper.
    [(lower), (upper)] - exclusive of both lower and upper values.

    The last case may also set using a simple two-value tuple:

    :example:
    >>> import probayes as pb
    >>> x = pb.Variable('x', vtype=float, vset=(1,2))
    >>> print(x.vset)
    [(1.0,), (2.0,)]
    """
    return self._vset

  @vset.setter
  def vset(self, vset=None):

    # Default vset to nominal
    if vset is None and self._vtype: 
      vset = DEFAULT_VSETS[self._vtype]
    elif isinstance(vset, (set, range)):
      vset = sorted(vset)
    elif np.isscalar(self._vset):
      vset = [self._vset]
    elif isinstance(vset, tuple):
      assert len(vset) == 2, \
          "Tuple vsets contain pairs of values, not {}".format(vset)
      vset = sorted(vset)
      vset = [(vset[0],), (vset[1],)]
    elif isinstance(vset, np.ndarray):
      vset = np.sort(vset).tolist()
    else:
      assert isinstance(vset, list), \
          "Unrecognised vset specification: {}".format(vset)

    # At this point, vset can only be a list, but may contain tuples
    vtype = self._vtype
    for i, value in enumerate(vset):
      if isinstance(value, tuple):
        if vtype is None:
          vtype = float
        val =  list(value)[0]
        if val != OO and val != -OO:
          vset[i] = ((vtype)(val),) 
      elif value == OO or value == -OO:
        if vtype is None:
          vtype = float
      else:
        if not vtype:
          vtype = eval_vtype(value)
        elif not self._vtype and vtype != eval_vtype(value):
          raise TypeError("Ambiguous value type {} vs {}".format(
            vtype, eval_vtype(value)))
        vset[i] = (vtype)(value)

    # Now vset contents should all be of the same vtype
    self._vset = vset
    vtype = eval_vtype(vtype)
    if self._vtype:
      assert self._vtype == vtype, \
          "Variable type {} incomptable with type for vset {}".format(
              self._vtype, vtype)
    else:
      self._vtype = vtype
    self._eval_vlims()

#-------------------------------------------------------------------------------
  @property
  def vtype(self):
    """ Property vtype is the variable type (default bool). If the variable set 
    if not set, then it is defaulted according to the variable type. """
    return self._vtype

  @vtype.setter
  def vtype(self, vtype=None):
    """ Sets the variable type (default bool). If the variable set if not set,
    then it is defaulted according to the variable type. """

    if not vtype:
      if self._vset is not None:
        self._vtype = eval_vtype(self._vset)
      else:
        self._vtype = DEFAULT_VTYPE
    else:
      self._vtype = eval_vtype(vtype)
    if self._vset is None:
      self.vset = DEFAULT_VSETS[self._vtype]

#-------------------------------------------------------------------------------
  @property
  def vlims(self):
    """ The untransformed (self._vlims) limits of the variable. """
    return self._vlims

  def _eval_vlims(self):
    """ Evaluates untransformed (self._vlims) and transformed (self._ulims) 

    :returns: the length of the variable.
    """
    self._vlims = None
    if self._vset is None:
      return self._eval_ulims()

    # Non-float limits are simple
    if self._vtype not in VTYPES[float]:
      self._vlims = np.array([min(self._vset), max(self._vset)])
      return self._eval_ulims()

    # Evaluates the limits from vset float
    assert len(self._vset) == 2, \
        "Floating point vset must be two elements, not {}".format(self._vset)
    lims = [None] * 2
    for i, limit in enumerate(self._vset):
        lim = limit if not isinstance(limit, tuple) else list(limit)[0]
        lims[i] = OO_TO_NP.get(lim, lim)

    # Re-order vset if limits are not sorted
    if lims[1] < lims[0]:
      vset = self._vset[::-1]
      self._vset = vset
    self._vlims = np.sort(lims)
    return self._eval_ulims()

#-------------------------------------------------------------------------------
  @property
  def ulims(self):
    """ The transformed (self._vlims) limits of the variable. """
    return self._ulims

  @property
  def inside(self):
    """ A lambda function for the whether an input is within the vset """
    return self._inside

  @property
  def isfinite(self):
    """ Boolean flag denoting whether variable range is finite """
    return self._isfinite

  @property
  def length(self):
    """ Length of the variable according to its (transformed) limits """
    return self._length

  @property
  def lhv(self):
    """ Log hypervolume (=log(length)) of the variable """
    return self._length

  def __len__(self):
    """ Length of the variable according to its (transformed) limits """
    return self._length

  def __contains__(self, *args, **kwds):
    if self._inside:
      return self._inside(*args, **kwds)
    return None

  def _eval_ulims(self):
    """ Evaluates transformed limits (self._ulims) and inside functions
    
    :returns: the length of the variable.
    """
    self._ulims = None
    self._length = None
    self._lhv = None
    self._inside = None
    self._isfinite = None
    if self._vlims is None:
      return self._length

    # Non-floats do not support transformation
    if self._vtype not in VTYPES[float]:
      self._inside = lambda x: np.isin(x, self._vset, assume_unique=True)
      self._ulims = self._vlims
      self._length = 2 if self._vtype in VTYPES[bool] else len(self._vset)
      self._lhv = log_prob(self._length)
      self._nlhv = -self._lhv
      self._isfinite = False
      return self._length

    # Floating point limits are subject to transformation
    self._ulims = self._vlims if self._ufun is None else \
                  self.ufun[0](self._vlims)
    self._isfinite = np.all(np.isfinite(self._ulims))
    self._length = max(self._ulims) - min(self._ulims)
    self._lhv = log_prob(self._length) if self._isfinite else np.inf
    self._nlhv = -self._lhv # This value serves as the coefficient for no_ucov

    # If not change of variables, restore all lims-related members except nlhv
    if self.__no_ucov:
      self._ulims = self._vlims
      self._isfinite = np.all(np.isfinite(self._ulims))
      self._length = max(self._ulims) - min(self._ulims)
      self._lhv = log_prob(self._length) if self._isfinite else np.inf

    # Now set inside function
    if not isinstance(self._vset[0], tuple) and \
        not isinstance(self._vset[1], tuple):
      self._inside = lambda x: np.logical_and(x >= self._vlims[0],
                                              x <= self._vlims[1])
    elif not isinstance(self._vset[0], tuple) and \
        isinstance(self._vset[1], tuple):
      self._inside = lambda x: np.logical_and(x >= self._vlims[0],
                                              x < self._vlims[1])
    elif isinstance(self._vset[0], tuple) and \
        not isinstance(self._vset[1], tuple):
      self._inside = lambda x: np.logical_and(x > self._vlims[0],
                                              x <= self._vlims[1])
    else:
      self._inside = lambda x: np.logical_and(x > self._vlims[0],
                                              x < self._vlims[1])

    return self._length

#-------------------------------------------------------------------------------
  @property
  def delta(self):
    """ Returns the default delta object if specified """
    return self._delta

  def set_delta(self, delta=None, *args, **kwds):
    """ Sets the default delta operation for the domain.

    :param delta: a callable or uncallable argument (see below)
    :param *args: args to pass if delta is callable.
    :param **kwds: kwds to pass if delta is callable (except scale and bound)

    The first argument delta may be:

    1. A callable function (operating on the first term).
    2. A Variable.Delta instance (this defaults all Variable deltas).
    3. A scalar that may or may not be contained in a container:
      a) No container - the scalar is treated as a fixed delta.
      b) List - delta is uniformly sampled from [-scalar to +scalar].
      c) Tuple - operation is +/-delta within the polarity randomised

    Two reserved keywords can be passed for specifying (default False):
      'scale': Flag to denote scaling deltas to Variable lengths
      'bound': Flag to constrain delta effects to Variable bounds
    """
    self._delta = delta
    self._delta_args = args
    self._delta_kwds = dict(kwds)
    if self._delta is None:
      return
    elif callable(self._delta):
      self._delta = Expression(self._delta, *args, **kwds)
      return

    # Default scale and bound
    if 'scale' not in self._delta_kwds:
      self._delta_kwds.update({'scale': False})
    if 'bound' not in self._delta_kwds:
      self._delta_kwds.update({'bound': False})

#-------------------------------------------------------------------------------
  @property
  def icon(self):
    return self._icon

  def set_icon(self, icon=None, *args, **kwds):
    """ Sets the variable symbol and carries members over to this Variable """
    if icon is None:
      icon = self._name
    """
    # If no arguments passed, default assumptions based on vtype and limits
    - unfortunately this messes up with subs(...)
    if isinstance(icon, str) and not args and not kwds:
      kwds = dict(**kwds)
      if self._vtype in VTYPES[float]:
        kwds.update({'integer': False, 'real': True})
        kwds.update({'finite': np.all(np.isfinite(self._vlims))})
        if np.max(self._vlims) > 0. and np.min(self._vlims) < 0.:
          pass
        elif np.min(self._vlims) >= 0.:
          kwds.update({'positive': True})
        elif np.max(self._vlims) <= 0:
          kwds.update({'negative': True})
        elif np.max(self._vlims) > 0.: 
          if isinstance(self._vset[0], tuple):
            kwds.update({'positive': True})
          else:
            kwds.update({'nonnegative': True})
        elif np.min(self._vlims) < 0. :
          if isinstance(self._vset[1], tuple):
            kwds.update({'negative': True})
          else:
            kwds.update({'nonpositive': True})
      elif self._vtype in VTYPES[int]:
        kwds.update({'integer': True})
        if np.max(self._vlims) > 0 and np.min(self._vlims) < 0:
          pass
        elif np.max(self._vlims) >= 0:
          kwds.update({'positive': True})
        elif np.min(self._vlims) <= 0:
          kwds.update({'negative': True})
    """
    return Icon.set_icon(self, icon, *args, **kwds)

#-------------------------------------------------------------------------------
  @property
  def ufun(self):
    return self._ufun

  @property
  def no_ucov(self):
    """ Flag to denote no univariate change of variables for iconic ufuns. """
    return self.__no_ucov


  def set_ufun(self, ufun=None, *args, **kwds):
    """ Sets a monotonic invertible tranformation for the domain as a tuple of
    two functions in the form (transforming_function, inverse_function) 
    operating on the first argument with optional further args and kwds.

    :param ufun: two-length tuple of monotonic functions.
    :param *args: args to pass to ufun functions.
    :param **kwds: kwds to pass to ufun functions.

    Support for this transformation is only valid for float-type vtypes.
    """
    self._ufun = ufun
    if self._ufun is None:
      return
    self.__no_ucov = False if 'no_ucov' not in kwds else kwds.pop('no_ucov')

    # Non-iconic ufun inputs must a two-length-tuple
    if not isiconic(self._ufun):
      assert self._vtype in VTYPES[float], \
          "Values transformation function only supported for floating point"
      message = "Non-iconic input ufun be a two-sized tuple of callable functions"
      assert isinstance(self._ufun, tuple), message
      assert len(self._ufun) == 2, message
      assert callable(self._ufun[0]), message
      assert callable(self._ufun[1]), message
    elif 'invertible' not in kwds:
      kwds.update({'invertible': True})
    self._ufun = Expression(self._ufun, *args, **kwds)
    if self.__no_ucov:
      assert self._ufun.isiconic, \
          "Can only set no_ucon=True for iconic univariate functions"
    self._eval_ulims()

#-------------------------------------------------------------------------------
  def evaluate(self, values=None):
    r""" Evaluates value(s) belonging to the variable.

    :param values: None, set of a single integer, array, or scalar.

    :return: a distribution dictionary {key: val} in which key is the variable 
             name and val is a NumPy array of values in accordance to the 
             following:

    If values is a NumPy array, it is returned unchanged.

    If values is None, it defaults to the entire variable set (vset) if not
    the variable type vtype is not float; otherwise a single scalar within the
    vset is randomly evaluated (see below for $n=0$).

    If values is a set containing a single integer (i.e. $\{n\}$), then the 
    output depends on the number $n$:

    If positive ($n$), then $n$ values are uniformly sampled.
    If zero ($n=0$), then a scalar value is randomly sampled.
    if negative($-n$), then $n$ values are randomly sampled.

    For non-float types, the values are evaluated from vset according to $n$:

    If positive ($n$), then $n$ values are serially sampled from ordered vset.
    If zero ($n=0$), then a scalar value is randomly sampled.
    if negative($-n$), then $n$ values are sampled from random vset permutations.
    
    For float types, then any uniformly sampled is performed in accordance of 
    any transformations set by Variable.set_ufun(), except for iconic ufun
    specifications with no change of variables (i.e. no_ucov=False).

    :example:
    >>> import numpy as np
    >>> import probayes as pb
    >>> freq = pb.Variable('freq', vtype=float, vset=[1., 8.])
    >>> freq.set_ufun((np.log, np.exp))
    >>> print(freq.evaluate({4}))
    freq: Distribution([('freq', array([1., 2., 4., 8.]))])
    """
    """ Evaluates the values ordered dictionary for __call__ """


    # If dictionary input type, values are keyed by variable name
    values = parse_as_str_dict(values) if isinstance(values, dict) \
             else {self.name: values}
    if isinstance(values, dict):
      values = values[self.name] 

    # Default to arrays of complete sets
    if values is None:
      if self._vtype in VTYPES[float]:
        values = {0}
      else:
        return Distribution(self._name, {self.name: 
                   np.array(list(self._vset), dtype=self._vtype)})

    # Sets may be used to sample from support sets
    if isunitset(values):
      number = list(values)[0]

      # Non-continuous
      if self._vtype not in VTYPES[float]:
        values = np.array(list(self._vset), dtype=self._vtype)
        if not number:
          values = values[np.random.randint(0, len(values))]
        else:
          if number > 0:
            indices = np.arange(number, dtype=int) % self._length
          else:
            indices = np.random.permutation(-number) % self._length
          values = values[indices]
        return Distribution(self._name, {self.name: values})
       
      # Continuous
      else:
        assert self._isfinite, \
            "Cannot evaluate {} values for bounds: {}".format(
                values, self._ulims)
        values = uniform(self._ulims[0], self._ulims[1], number, 
                           isinstance(self._vset[0], tuple), 
                           isinstance(self._vset[1], tuple)
                        )

      # Only use ufun when isunitsetint(values)
      if self._ufun and not self.__no_ucov:
        return Distribution(self._name, {self.name: self.ufun[-1](values)})
    return Distribution(self._name, {self.name: values})

#-------------------------------------------------------------------------------
  def __call__(self, values=None):
    """ See Variable.evaluate() """
    return self.evaluate(values)

#-------------------------------------------------------------------------------
  def __repr__(self):
    """ Printable representation of variable including name """
    return self._name

#------------------------------------------------------------------------------- 
  def eval_delta(self, delta=None):
    """ Evaluates the value(s) of a delta operation without applying them.

    :param delta: delta value(s) to offset (see Variable.apply_delta).
    :return: the evaluated delta offset values.
    :rtype Variable.delta()

    If delta is not entered, then the default set by Variable.set_delta() is used.
    """
    delta = delta or self._delta
    if delta is None:
      return None
    if isinstance(delta, Expression):
      if delta.ret_callable():
        return delta
      delta = delta()
    if isinstance(delta, self._Delta):
      delta = delta[0]
    orand = isinstance(delta, tuple)
    urand = isinstance(delta, list)
    if orand:
      assert len(delta) == 1, "Tuple delta must contain one element"
      delta = delta[0]
      if self._vtype not in VTYPES[bool]:
        delta = delta if np.random.uniform() > 0.5 else -delta
    elif urand:
      assert len(delta) == 1, "List delta must contain one element"
      delta = delta[0]
      if self._vtype in VTYPES[bool]:
        pass
      elif self._vtype in VTYPES[int]:
        delta = np.random.randint(-delta, delta)
      else:
        delta = np.random.uniform(-delta, delta)
    assert isscalar(delta), "Unrecognised delta type: {}".format(delta)
    if delta == self._delta and self._delta_kwds['scale']:
      assert np.isfinite(self._length), "Cannot scale by infinite length"
      delta *= self._length
    return self._Delta(delta)

#------------------------------------------------------------------------------- 
  def apply_delta(self, values, delta=None, bound=None):
    """ Applies delta operation  to values optionally contrained by bounds.

    :param values: Numpy array values to apply.
    :param delta: delta value(s) to offset to the values
    :param bound: optional argument to contrain outputs.

    :return: Returns the values following the delta operation.

    If delta is not entered, then the default set by Variable.set_delta() is used.
    Delta may be a scalar or a single scalar value contained in a tuple or list.

    1. A scalar value: is summated to values (transformed if ufun is specified).
    2. A tuple: the polarity of the scalar value is randomised for the delta.
    3. A list: the delta is uniformly sampled in the range [0, scalar].
    """

    # If dictionary input type, values are keyed by variable name
    if isinstance(values, dict):
      values = values[self.name] 

    # Call eval_delta() if values is a list and return values if delta is None
    delta = delta or self._delta
    if isinstance(delta, Expression):
      if delta.ret_callable():
        return delta(values)
      delta = delta()
    elif self._vtype not in VTYPES[bool]:
      if isinstance(delta, (list, tuple)):
        delta = self.eval_delta(delta)
    if isinstance(delta, self._Delta):
      delta = delta[0]
    if delta is None:
      return values

    # Apply the delta, treating bool as a special case
    if self._vtype in VTYPES[bool]:
      orand = isinstance(delta, tuple)
      urand = isinstance(delta, list)
      if orand or urand:
        assert len(delta) == 1, "Tuple/list delta must contain one element"
        delta = delta[0]
        if isscalar(values) or orand:
          vals = values if delta > np.random.uniform() > 0.5 \
                 else np.logical_not(values)
        else:
          flip = delta > np.random.uniform(size=values.shape)
          vals = np.copy(values)
          vals[flip] = np.logical_not(vals[flip])
      else:
        vals = np.array(values, dtype=int) + np.array(delta, dtype=int)
        vals = np.array(np.mod(vals, 2), dtype=bool)
    elif self._ufun is None or self.__no_ucov:
      vals = values + delta
    else:
      transformed_vals = self.ufun[0](values) + delta
      vals = self.ufun[1](transformed_vals)
    vals = revtype(vals, self._vtype)

    # Apply bounds
    if bound is None:
      bound = False if 'bound' not in self._delta_kwds \
             else self._delta_kwds['bound']
    if not bound:
      return vals
    maybe_bounce = [False] if self._vtype not in VTYPES[float] else \
                   [isinstance(self._vset[0], tuple), 
                    isinstance(self._vset[1], tuple)]
    if not any(maybe_bounce):
      return np.maximum(self._vlims[0], np.minimum(self._vlims[1], vals))

    # Bouncing scalars and arrays without and with boolean indexing respectively
    if isscalar(vals):
      if all(maybe_bounce):
        if not self._inside(vals):
          vals = values
      elif maybe_bounce[0]:
        if vals < self._vlims[0]:
          vals = values
        else:
          vals = np.minimum(self._vlims[1], vals)
      else:
        if vals > self._vlims[1]:
          vals = values
        else:
          vals = np.maximum(self._vlims[0], vals)
    else:
      if all(maybe_bounce):
        outside = np.logical_not(self._inside(vals))
        vals[outside] = values[outside]
      elif maybe_bounce[0]:
        outside = vals <= self._vlims[0]
        vals[outside] = values[outside]
        vals = np.minimum(self._vlims[1], vals)
      else:
        outside = vals >= self._vlims[1]
        vals[outside] = values[outside]
        vals = np.maximum(self._vlims[0], vals)
    return vals

#-------------------------------------------------------------------------------
  def inverse(self):
    """ Returns the icon unless an invertible Sympy ufun has been set, in which
    case the corresponding inverse icon is returned.
    """
    if self._ufun and self._ufun.inverse is not None:
      return self._ufun.inverse
    return self._icon

#-------------------------------------------------------------------------------
  def __invert__(self):
    """ See Variable.inverse() """
    return self.inverse()

#-------------------------------------------------------------------------------
  @property
  def dexpr(self):
    """ Sets iconic deterministic expression, substituting for icon in [:] """
    return self._dexpr

  @name.setter
  def dexpr(self, dexpr=None):
    self._dexpr = None if self._is_stochastic else dexpr

#-------------------------------------------------------------------------------
  def __getitem__(self, arg):
    """ Method [:] overloaded to return icon or dexpr object. """
    assert arg == slice(None), \
        "Only ':' input accepted for __getitem__ method  Icon[:], not: {}".\
        format(arg)
    return self._dexpr or self._icon

#-------------------------------------------------------------------------------
  def __and__(self, other):
    """ Combination operator between Variable and another Variable, Field, or
    Dependence. """
    return and_op(self, other)

#-------------------------------------------------------------------------------
  def __or__(self, other):
    """ Combination operator between Variable and another Variable, Field, or
    Dependence. """
    return or_op(self, other)

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
def variables(names, vtype=None, vset=None, *args, **kwds):
  if isinstance(names, str):
    names = names.split(',')
  varlist = [Variable(name, vtype, vset, *args, **kwds) for name in names]
  return tuple(varlist)

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
if __name__ == "__main__":
  import doctest
  doctest.testmod()

#-------------------------------------------------------------------------------



""" Symbolic representation wrapping SymPy's Symbol and Expr. """

#-------------------------------------------------------------------------------
import sympy
import sympy.stats

#-------------------------------------------------------------------------------
def isiconic(var):
  """ Returns whether object is a Sympy object. This function is purposefull not
  called issymbolic because he use 'icon' to denote any symbolic object]
  including sympy Integer and Float types. """
  # Non-callables objects derive from sympy.Basic
  if not callable(var):
    return isinstance(var, sympy.Basic)
  return False

  """
  # A hacky solution for functions
  import sys
  module_path = str(sys.modules.get(var.__module__))
  return 'sympy.' in module_path and '/sympy/' in module_path
  """

#-------------------------------------------------------------------------------
class Icon:
  """ This class wraps sympy.Symbol and sympy.Expr denoting an 'icon' as any
  object derived from sym
  . Sympy's dependence on __new__ 
  to return modified class objects at instantiation is makes multiple
  inheritance tricky so instead we wrap them in here as a class and copy over 
  the attributes.

  The resulting instance can be treated as a SymPy object using the __invert__
  method (~instance):

  :example
  >>> import probayes as pb
  >>> x = pb.Icon('x')
  >>> x2 = 2 * x[:]
  >>> print(x2.subs({x[:]: 4}))
  8
  >>>
  """
  _icon = None

#-------------------------------------------------------------------------------
  def __init__(self, icon, *args, **kwds):
    self.set_icon(icon, *args, **kwds)

#-------------------------------------------------------------------------------
  @property
  def icon(self):
    return self._icon

  def set_icon(self, icon, *args, **kwds):
    """ Sets the icon object for this instance with optional args and kwds.
    Either pass a Sympy expression or sympy.Symbol object directly or in accordance 
    with the calling conventions for sympy.Symbol.__new__
    """

    # Pass symbolic or create symbolc named according to string
    self._icon = icon
    if isiconic(self._icon):
      pass
    elif isinstance(self._icon, str):
      self._icon = sympy.Symbol(self._icon, *args, **kwds)
    else:
      raise TypeError("Symbol name must be string; {} entered".format(self._icon))

    # Copy attributes and hash content
    members = dir(self._icon)
    for member in members:
      if not hasattr(self, member):
        try:
          attribute = getattr(self._icon, member)
          setattr(self, member, attribute)
        except AttributeError:
          pass

#-------------------------------------------------------------------------------
  def __repr__(self):
    if self._icon is None:
      return super().__repr__()
    return self._icon.__repr__()

#-------------------------------------------------------------------------------
  def __hash__(self):
    if self._icon is None:
      return super().__hash__()
    return self._icon.__hash__()

#-------------------------------------------------------------------------------
  def __getitem__(self, arg):
    """ Method [:] overloaded to return icon object. """
    assert arg == slice(None), \
        "Only ':' input accepted for __getitem__ method  Icon[:], not: {}".\
        format(arg)
    return self._icon

#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
#-------------------------------------------------------------------------------
if __name__ == "__main__":
  import doctest
  doctest.testmod()

#-------------------------------------------------------------------------------



'''
A wrapper for expression functions that makes optional use of 'order' and 'delta' 
specifications. Multiple outputs are supported for non-symbolic expressions. Note
that an expression may comprise a constant by not being callable, and therefore 
not all expressions are functions.
'''
import types
import collections
import functools
import sympy
from probayes.vtypes import isscalar
from probayes.icon import isiconic
from probayes.expr import Expr

#-------------------------------------------------------------------------------
def islambda(func):
  if not callable(func):
    return False
  if type(func) is types.LambdaType:
    return func.__name__ == '<lambda>'

#-------------------------------------------------------------------------------
class Expression:
  """ A expression wrapper to enable object representations as an uncallable
  array, a callable function, or a tuple/dict of callable functions

  :example:
  >>> from probayes.expression import Expression
  >>> hw = Expression("Hello World!")
  >>> print(hw())
  Hello World!
  >>> inc = Expression(lambda x: x+1)
  >>> print(inc(2.))
  3.0
  >>> inc_dec = Expression( (lambda x:x+1, lambda x:x-1) )
  >>> print(inc_dec[0](3.))
  4.0
  >>> print(inc_dec[1](3.))
  2.0
  >>> sqr_sqrt = Expression( {'sqr': lambda x:x**2, 'sqrt': lambda x:x**0.5} )
  >>> print(sqr_sqrt['sqr'](3.))
  9.0
  >>> print(sqr_sqrt['sqrt'](4.))
  2.0

  Since multiple expressions are supported, the symbolic value of Expression
  is not related to the input.
  """

  # Protected
  _expr = None     # Expression set
  _args = None     # Optional args
  _kwds = None     # Optional kwds
  _exprs = None    # Dict of iconic Expr objects
  _partials = None # Dict of partial functions
  _keys = None     # Keys of partials
  _ismulti = None  # Flag to denote multiple expression
  _callable = None # Flag to denote callable expression
  _islambda = None # Flag to denote lambda callable
  _isscalar = None # Flag to denote uncallable scalar
  _isiconic = None # Flag to denote sympy expression
  _symbols = None  # Dict of iconic symbols used

  # Private
  __invertible = None
  __inverse = None
  __invexpr = None
  __derinv = None # derivative of value with respect to inverse
  __order = None
  __delta = None

#-------------------------------------------------------------------------------
  def __init__(self, expr=None, *args, **kwds):
    """ Initialises instances according to object in expr, which may be an 
    uncallable object, a callable function, or a tuple of callable functions. 
    See set_expr()
    """
    self.set_expr(expr, *args, **kwds)
    
#-------------------------------------------------------------------------------
  @property
  def expr(self):
    return self._expr

  @property
  def args(self):
    return self._args

  @property
  def kwds(self):
    return self._kwds

  @property
  def order(self):
    return self.__order

  @property
  def delta(self):
    return self.__delta

  @property
  def callable(self):
    return self._callable

  @property
  def islambda(self):
    return self._islambda

  @property
  def ismulti(self):
    return self._ismulti

  @property
  def isscalar(self):
    return self._isscalar

  @property
  def isiconic(self):
    return self._isiconic

  @property
  def inverse(self):
    return self.__inverse

  @property
  def invexpr(self):
    return self.__invexpr

  @property
  def derinv(self):
    return self.__derinv

  @property
  def invertible(self):
    return self.__invertible

  @property
  def keys(self):
    return self._keys

  @property
  def symbols(self):
    return self._symbols

  def set_expr(self, expr=None, *args, **kwds):
    """ Set the Func instance's function object.

    :param expr: an uncallable object, callable function, or tuple of functions
    :param *args: arguments to pass onto callables
    :param **kwds: keywords to pass onto callables

    Note that the following two reserved keywords are disallowed:

    'order': which instead denotes a dictionary of remappings.
    'delta': which instead denotes a mapping of differences.
    """
    self._expr = expr
    self._args = tuple(args)
    self._kwds = dict(kwds)
    self._callable = None
    self._islambda = None
    self._ismulti = None
    self._isscalar = None
    self._isiconic = None
    self.__order = None
    self.__delta = None
    self.__inverse = None
    self.__invexpr = None
    self.__invderi = None
    self.__invertible = False if 'invertible' not in kwds else \
                        self._kwds.pop('invertible')
    self._exprs = collections.OrderedDict()
    self._symbols = collections.OrderedDict()

    # Sanity check func
    if self._expr is None:
      assert not args and not kwds, "No optional args without a function"
    self._isscalar = isscalar(self._expr)
    self._isiconic = isiconic(self._expr)
    self._ismulti = isinstance(self._expr, (dict, tuple))
    if self.__invertible:
      assert not self._ismulti,\
          "Cannot invert for multiple expressions"

    # Non-multi iconic
    if self._isiconic:
      self._exprs.update({None: Expr(self._expr, *self._args, **self._kwds)})
      self._symbols.update(self._exprs[None].symbols)
      self._ismulti = self.__invertible and \
                      len(self._exprs[None].symbols) == 1
      self._callable = True
      self._islambda = False

    # Unitary
    elif not self._ismulti:
      self._callable = callable(self._expr)
      self._islambda = islambda(self._expr)
      if not self._callable:
        assert not args and not kwds, \
            "No optional arguments with uncallable expressions"
        self._isscalar = isscalar(self._expr)

    # Multi
    else:
      exprs = self._expr if isinstance(self._expr, tuple) else \
              self._expr.values()
      self._callable = False
      self._islambda = False
      self._isscalar = False
      self._isiconic = False
      each_callable = [callable(expr) for expr in exprs]
      each_islambda = [islambda(expr) for expr in exprs]
      each_isscalar = [isscalar(expr) for expr in exprs]
      each_isiconic = [isiconic(expr) for expr in exprs]
      assert len(set(each_callable)) < 2, \
          "Cannot mix callable and uncallable expressions"
      assert len(set(each_islambda)) < 2, \
          "Cannot mix lambda and non-lambda expressions"
      assert len(set(each_isscalar)) < 2, \
          "Cannot mix scalars and nonscalars"
      assert len(set(each_isiconic)) < 2, \
          "Cannot mix iconics and non-iconics"
      if len(self._expr):
        self._callable = each_callable[0]
        self._islambda = each_islambda[0]
        self._isscalar = each_isscalar[0]
        self._isiconic = each_isiconic[0]
      if not self._callable:
        assert not args and not kwds, "No optional args with uncallable function"
      if self._isiconic:
        assert not args and not kwds, "No optional args with iconic function"
      if self._isiconic:
        if isinstance(self._expr, tuple):
          for i, expr in enumerate(self._expr):
            self._exprs.update({i: Expr(expr, *self._args, **self._kwds)})
            self._symbols.update(self._exprs[i].symbols)
        elif isinstance(self._expr, dict):
          for key, val in self._expr.items():
            self._exprs.update({key: Expr(val, *self._args, **self._kwds)})
            self._symbols.update(self._exprs[key].symbols)
    if 'order' in self._kwds:
      self.set_order(self._kwds.pop('order'))
    if 'delta' in self._kwds:
      self.set_delta(self._kwds.pop('delta'))
    self._set_partials()
    self._keys = list(self._partials.keys())

#-------------------------------------------------------------------------------
  def set_order(self, order=None):
    """ Sets an order remapping dictionary for functional calls in which
    keyword arguments are mapped to position (in numeric) or rekeyed (if str).
    """
    self.__order = order
    if self.__order is None:
      return
    assert self.__delta is None, "Cannot set both order and delta"
    self._check_mapping(self.__order)

#-------------------------------------------------------------------------------
  def set_delta(self, delta=None):
    """ Sets a difference remapping dictionary for functional calls in which
    keyword arguments are mapped to position (in numeric) or rekeyed (if str).
    """
    self.__delta = delta
    if self.__delta is None:
      return
    assert self.__order is None, "Cannot set both order and delta"
    assert not self.isiconic, "Cannot set delta when using symbols"
    self._check_mapping(self.__delta)

#-------------------------------------------------------------------------------
  def _check_mapping(self, mapping=None):
    """ Perform sanity checkings on mapping dictionary """
    if mapping is None:
      return
    # Used to sanity-check mapping dicts e.g. order and delta
    assert isinstance(mapping, dict), \
        "Mapping must be a dictionary type, not {}".format(type(mapping))
    if self.isiconic:
      assert not any(list(mapping.values())), \
          "Cannot remap for iconic expressions"
    key_list = list(mapping.keys())
    ind_list = list(mapping.values())
    keys = []
    inds = []
    for key, ind in zip(key_list, ind_list):
      keys.append(key)
      if type(ind) is int:
        inds.append(ind)
      elif ind is None:
        pass
      elif not isinstance(ind, str):
        raise TypeError("Cannot interpret index specification value: {}".ind)
    indset = set(inds)
    if len(indset):
      assert indset == set(range(min(indset), max(indset)+1)), \
          "Index specification non_sequitur: {}".format(indset)

#-------------------------------------------------------------------------------
  @property
  def partials(self):
    return self._partials

  def _set_partials(self):
    # Protected function to update partial function dictionary of calls
    self._partials = collections.OrderedDict()

    # Iconic partials calls are set to Expr.__call__
    if self._isiconic:
      if not self._ismulti or self.__invertible:
        call = functools.partial(Expr.__call__, self._exprs[None])
        self._partials.update({None: call})

        # If invertible, solve the expression
        if self.__invertible:
          self._partials.update({0: call})
          expr = self._exprs[None]
          key = list(expr.symbols.keys())[0]
          derinv = sympy.Derivative(expr, self._exprs[None].symbols[key]).doit()
          self.__derinv = Expr(derinv)
          inv_key = "_{}_inv".format(key)
          self.__inverse = sympy.Symbol(inv_key)
          invexprs = sympy.solve(self._expr - self.__inverse, self._exprs[None].symbols[key])
          n_exprs = len(invexprs)
          assert n_exprs, "No invertible solutions for expression {}".format(
              self._expr)
          invexpr = invexprs[0]
          if n_exprs > 1:
            for expr in invexprs[1:]:
              if len(expr.__repr__()) < len(invexpr.__repr__()):
                invexpr = expr
          self.__invexpr = Expr(invexpr)
          call = functools.partial(Expr.__call__, self.__invexpr)
          self._partials.update({1: call})
          self._partials.update({-1: call})
      else:
        if isinstance(self._exprs, tuple):
          for i, expr in enumerate(self._exprs):
            call = functools.partial(Expr.__call__, expr)
            self._partials.update({i: call})
          self._partials.update({-1: call})
        elif isinstance(self._exprs, dict):
          for key, val in self._exprs.items():
            call = functools.partial(Expr.__call__, val)
            self._partials.update({key: call})

    # Non-multiples are keyed by: None
    elif not self._ismulti:
      if self._isiconic:
        call = functools.partial(Expr.__call__, self)
        self._partials.update({None: call})
      else:
        call = functools.partial(Expression._partial_call, self, self._expr, 
                                 *self._args, **self._kwds)
        self._partials.update({None: call})

    # Tuples are keyed by index
    elif isinstance(self._expr, tuple):
      for i, expr in enumerate(self._expr):
        call = functools.partial(Expression._partial_call, self, expr, 
                                 *self._args, **self._kwds)
        self._partials.update({i: call})
      self._partials.update({-1: call})

    # Dictionaries keys are mapped directly
    elif isinstance(self._expr, dict):
      for key, val in self._expr.items():
        call = functools.partial(Expression._partial_call, self, val, 
                                 *self._args, **self._kwds)
        self._partials.update({key: call})
    else:
      raise TypeError("Unknown expression type: {}".format(type(self._expr)))
    self._keys = list(self._partials.keys())

#-------------------------------------------------------------------------------
  def _call(self, expr=None, *args, **kwds):
    """ Private call used by the wrapped Func interface that is _ismulti-blind.
    (see __call__ and __getitem__).
    """
    assert not self._isiconic, \
        "Iconic calls must be handled by Expr() objects"
    #argsin = args; kwdsin = kwds; import pdb; pdb.set_trace() # debugging

    # Non-callables
    if not self._callable:
      assert not args and not kwds, \
          "No optional args with uncallable or symbolic expressions"
      if self._ismulti:
        return expr
      if expr is None:
        expr = self._expr
      return expr

    # Allow first argument to denote kwds
    if len(args) == 1 and isinstance(args[0], dict):
      args, kwds = (), {**kwds, **dict(args[0])}
    #argsmid = args; kwdsmid = kwds; import pdb; pdb.set_trace() # debugging

    # Callables with neither order nor delta are straightforwards
    if not self.__order and not self.__delta:
      if self._islambda and hasattr(expr, '__code__') and \
          not expr.__code__.co_argcount: # Allow argument-free lambdas
        assert not len(args), \
            "Unexpected entering of positional arguments: {}".format(args)
        return expr()
      elif args or all(isinstance(key, str) for key in kwds.keys()):
        return expr(*args, **kwds)
      else:
        return expr(dict(kwds))

    # Append None to args according to mapping index specification
    n_args = len(args)
    mapping = self.__order or self.__delta
    for val in mapping.values():
      if type(val) is int:
        n_args = max(n_args, val+1)
    args = list(args)
    while len(args) < n_args:
      args.append(None)

    # Callables with order wrapper
    if self.__order:
      for key, val in self.__order.items():
        if type(val) is int:
          args[val] = kwds.pop(key)
        elif val is None:
          kwds.pop(key)
        elif isinstance(val, str):
          kwds.update({val: kwds.pop(key)})
        else:
          raise TypeError("Unrecognised order key: val type: {}:{}".\
                          format(key, val))
      return expr(*tuple(args), **kwds)

    # Callables with delta wrapper
    for key, val in self.__delta.items():
      if key[-1] != "'":
        value = kwds.pop(key)
      else:
        value = kwds.pop(key) - kwds.pop(key[:-1])
      if type(val) is int:
        args[val] = value
      elif val is None:
        pass
      elif isinstance(val, str):
        kwds.update({val: value})
      else:
        raise TypeError("Unrecognised delta key: val type: {}:{}".\
                        format(key, val))

    #argsout = args; kwdsout = kwds; import pdb; pdb.set_trace() # debugging
    return expr(*tuple(args), **kwds)

#-------------------------------------------------------------------------------
  def _partial_call(self, *args, **kwds):
    return self._call(*args, **kwds)

#-------------------------------------------------------------------------------
  def __call__(self, *args, **kwds):
   """ Wrapper call to the function with optional inclusion of additional
   args and kwds. """
   assert not self._ismulti, \
       "Cannot call with multiple expression, use Expression[{}]()".format(
           list(self._keys()))
   return self._partials[None](*args, **kwds)

#-------------------------------------------------------------------------------
  def __getitem__(self, arg):
   r""" Returns the $i$th function from the expr tuple where if is $i$ is
   numeric, otherwise arg is treated as key returning the corresponding
   value in the expr dictionary. """
   if arg is not None: 
     assert self._ismulti, \
         "Cannot call with non-multiple expression, use Expression()"
   if isinstance(arg, slice):
     if arg == slice(None):
       return self._expr
   if isinstance(arg, (tuple, list)) and not len(arg):
     if isinstance(arg, tuple):
       return self._partials
     return self._partials[None]
   return self._partials[arg]

#-------------------------------------------------------------------------------
  def __len__(self):
    """ Returns the number of expressions in the tuple set by set_expr() """
    if not self._ismulti:
      return None
    return len(self._partials)

#-------------------------------------------------------------------------------
  def __repr__(self):
    """ Print representation """
    if self._isiconic or not self._callable:
      return object.__repr__(self)+ ": '{}'".format(self._expr)
    if self._keys is None:
      return object.__repr__(self) 
    return object.__repr__(self)+ ": '{}'".format(self._keys)

#-------------------------------------------------------------------------------

#-------------------------------------------------------------------------------
if __name__ == "__main__":
  import doctest
  doctest.testmod()

#-------------------------------------------------------------------------------



'''
Utility module for variable.Variable() (and therefore many other objects)
'''
import collections

#-------------------------------------------------------------------------------
def parse_as_str_dict(*args, **kwds):
  """ Aggregrates all arguments in args (which must be dictionary objects) and
  keywords to output as a single ordereddict(), ensuring all keys are strings,
  replacing all non-string keys with arg.name values. Duplicate keys are not
  checked. Note 'remap' is a reserved keyword to allow key substituion using
  a dictionary. """

  remap = None if 'remap' not in kwds else kwds.pop('remap')
  kwds = collections.OrderedDict(kwds)
  if not args:
    return kwds
  args_dict = collections.OrderedDict()
  for arg in args:
    assert isinstance(arg, dict), \
        "Each argument type must be dict, not {}".format(type(arg))
    if not remap and all([isinstance(key, str) for key in arg.keys()]):
      args_dict.update(arg)
    else:
      for key, val in arg.items():
        if isinstance(key, str):
          if remap and key in remap:
            args_dict.update({remap[key]: val})
          else:
            args_dict.update({key: val})
        else:
          keys = key if isinstance(key, (list,tuple)) else [key]
          key_name = [None] * len(keys)
          for i, subkey in enumerate(keys):
            if isinstance(subkey, str):
              key_name[i] = subkey
            elif hasattr(subkey, 'name'):
              key_name[i] = subkey.name
              assert isinstance(key_name[i], str), \
                  "Non-string name attribute {} for key object: {}".format(
                      key_name[i], key)
            else:
              raise KeyError("Uninterpretable key to parse: {}".format(key))
          key_name = ','.join(key_name)
          if remap and key_name in remap:
            args_dict.update({remap[key_name]: val})
          else:
            args_dict.update({key_name: val})
  if kwds:
    args_dict.update(kwds)
  return args_dict

#-------------------------------------------------------------------------------
